#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include "Assm.h"

//#define MESSAGE 	1
//#define WAIT 		1

CAssemble::CAssemble()
{
	Warnings=Errors=0;
	fptr=NULL;
	fpList=NULL;
}

CAssemble::CAssemble(char * file)
{
	char *c;
	CString Source;
	Source = file;
	Name = file;
	m_ObjFile = NULL;
	if( ( c=strchr(file,(int)'.') )!=NULL)
	{
		*c=0;
	}
	else
		Source += ".asm";

	ListFile = file;
	ObjectFile = file;
	ListFile += ".lst";
	ObjectFile += ".obj";
	Warnings=Errors=0;

	fptr=Source.FileOpen("r");
	fpList=ListFile.FileOpen("w");
}

CAssemble::~CAssemble()
{
	fclose(fptr);
	fclose(fpList);
}

int CAssemble::Pass1()
{
	// *********** local variables ***********
	CString temp;  	//temporary string
	int ret;			//return code of the CSymLit::AddItem() function
	int found; 			//set to 1 when a op code is found
	int flag;       	//1-pop 	2-mop
	int q;				//used to hold length
	List Entry;     	//variable to hold details before it is put in a CSymLit
	MOT *pmot;			//pointer to a MOT Entry

	// *********** Pass 1 ***********
	Done=0;
	LC=0;
	if(fptr==NULL)
	{
		printf("\nUnable to open source file");
		return 0;
	}
	LabelLength = sizeof(Entry.Segment);
	#ifdef MESSAGE
	printf("\nPass1...");
	#endif
	fseek(fptr,0,SEEK_SET);
	LineNo=0;

	while( String.ReadFile(fptr) && !Done)
	{
		String.Trim();
//		String.ToUpper();
		found=0;
		LineNo++;
		Read=GetNextToken(String,temp,0);
		if(!Read)
			continue;
		Label.Empty();
		flag=2;
		q = temp.GetLength() - 1;
		if(temp[q] == ':')//Label
		{
			temp.SetAt(q,0);
			if( CanBeLabel(temp) )
			{
				Label=temp;
				Label.SetLength( LabelLength );
			}
			else
			{
				Error("\nInvalid instruction/label");
				continue;
			}
		}
		else if((idx=GetPOT(temp))!=-1)    //Pseudo op
		{
			flag=1;
			found=1;
		}
		else if((idx=GetMOT(temp))!=-1) //Machine op
			found=1;
		if(!found)
		{
			if( CanBeLabel(temp) )
			{
				Label=temp;
				Label.SetLength(LabelLength);
			}
			else
			{
				Warning("Invalid label; ignored");
				temp.Print();
			}
			Read+=GetNextToken(String,temp,Read);
			if((idx=GetPOT(temp))!=-1)    //Pseudo op
			{
				flag=1;
				found=1;
			}
			else if((idx=GetMOT(temp))!=-1) //Machine op
				found=1;

		}
		if(!found)//still no op code found-
		{
			Error("Invalid Mnemonic ");
			temp.Print();
			continue;
		}
		mne_op=temp;

		if(flag == 2 && Label.GetLength())//Machine op label exists
		{
			Label.CopyTo(Entry.symbol);
			Entry.value=LC;
			Entry.CombineClass = ' ';
			CurrentSegment.CopyTo(Entry.Segment);
			Entry.Relocation='R';
			Entry.Type='N';
			ret=Sym.AddItem(Entry);
			if(ret==-1)
			{
				Error(" Symbol table full. Insufficient memory!");
			}
		}
		if(flag==2)	//Machine op
		{
			ret=GetOperands();
			if( ret == 0 ) 		//no operands
			{
				while(idx!=-1)
				{
					pmot=Tables.GetMOTAt(idx);
					if( (pmot->type==0 && pmot->Accumulator==0
							&& pmot->SegmentRegister==0 && pmot->Reg_op==0 )
								|| pmot->type == 1)
						break;
					idx=GetNextMOT(temp,idx);
				}
				if( idx == -1 )
				{
					Error ("No operands specified");
					continue;
				}

				LC++;
				if(pmot->type==1)
					LC++;
			}
			else //one or two operand
			{
				while(idx!=-1)
				{
					pmot=Tables.GetMOTAt(idx);
					ret=CheckType(pmot,1);
					if (ret == -1 )
						break;
					else if( ret !=0 )
					{
						LC += ret;
						break;
					}
					idx=GetNextMOT(temp,idx);
				}
				if( ret == -1 )
					continue;
				if(idx == -1 )
				{
					Error("Unable to handle mnemonic.");
					continue;
				}
			}
		}
		else //Pseudo op
			HandlePseudoOP(idx,1);
	}//end of while
	int error;
	if( ! Sym.SetPublics(temp, error) )
	{
		if(error == 1)
			Error("Public variable not defined");
		else if(error == 2)
			Error("Variable already extern; cannot be made public");
		temp.Print(" - %s");
	}
	if(Errors)
		return 0;
	else
		return 1;

}

int CAssemble::Pass2()
{
	// *********** local variables ***********

	MOT *pmot;		//pointer to a MOT entry
	int ret;		//return code of the CSymLit::AddItem() function
	int found; 		//set to 1 when a op code is found
	int flag;  		//1-pop 	2-mop
	int i;
	CString temp;  	//temporary string
	List *Item;
	CObjFile ObjFile( ObjectFile , Name);
	m_ObjFile = &ObjFile;
	// *********** Pass 2 ***********
	LC=0;
	Done=0;
	#ifdef MESSAGE
	printf("\nPass2...");
	#endif
	if(fptr==NULL)
		return 0;
	fseek(fptr,0,SEEK_SET);
	LineNo=0;

	//Writing LNAMES Record
	ObjFile.WriteLNameHdr();
	ObjFile.WriteLName( CString() );
	for( i = 0, idx = 1 ; ; i++ )
	{
		Item=Sym.GetAt( i );
		if( Item == NULL )
			break;
		if( Item->Type == 'S')
		{
			idx++;
			ObjFile.WriteLName( CString( Item->symbol) );
			Item->LnameIndex = idx;
		}
	}
	ObjFile.WriteLName(); //Done writing LNAMES

	//SEGDEF Records

	BYTE ACBP;
	WORD SegLength;
	BYTE Length;
	BYTE TypeIndex = 0;
	WORD SegmentIndex = 1;

	for( i = 0, idx = 1 ; ; i++ )
	{
		Item=Sym.GetAt( i );
		if( Item == NULL )
			break;
		if( Item->Type == 'S')
		{
			Item->Index = SegmentIndex;
			Item->GroupIndex = 0;

			ACBP = Item->Align << 5;
			if( Item->CombineClass == 'P')
				ACBP |= 8;
			else if( Item->CombineClass == 'S')
				ACBP |= 20;
			else if( Item->CombineClass == 'C')
				ACBP |= 24;
			SegLength = Item->SegmentLength;

			// Write SEGDEF Record
			m_ObjFile->WriteRecHdr( 0x98 );
			m_ObjFile->WriteRec( &ACBP, 1 );
			m_ObjFile->WriteRec( (BYTE *) &SegLength, 2);
			m_ObjFile->WriteIndex( Item->LnameIndex );
			m_ObjFile->WriteIndex( 1 ); //class index
			m_ObjFile->WriteIndex( 1 ); //overlay index
			m_ObjFile->WriteChksum();
			SegmentIndex ++;
		}
	}
	//SEGDEF done

	//EXTDEF PUBDEF
	WORD ExtIndex = 1;
	WORD PubIndex = 1;

	for( i = 0, idx = 1 ; ; i++ )
	{

		Item=Sym.GetAt( i );
		if( Item == NULL )
			break;
		if( Item->Type == 'S'); //Ignore segments
		else if ( Item->CombineClass =='E' )
		{
			Item->Index = ExtIndex ++;
			m_ObjFile->WriteRecHdr( 0x8C );
			Length = strlen(Item->symbol);
			m_ObjFile->WriteRec(&Length, 1);
			m_ObjFile->WriteRec(Item->symbol, Length);
			m_ObjFile->WriteRec( &TypeIndex, 1 ); //type index
			m_ObjFile->WriteChksum();
		}
		else if ( Item->CombineClass =='P' )
		{
			Item->Index = PubIndex++;
			m_ObjFile->WriteRecHdr( 0x90 );
			Length = strlen(Item->symbol);
			List *Item1 = Sym.GetItem( CString( Item->Segment ) );
			m_ObjFile->WriteIndex( Item1->GroupIndex);
			m_ObjFile->WriteIndex( Item1->Index);
			m_ObjFile->WriteRec(&Length, 1);
			m_ObjFile->WriteRec(Item->symbol, Length);
			m_ObjFile->WriteRec( (BYTE *) &Item->value, 2);
			m_ObjFile->WriteRec( &TypeIndex, 1 ); //type index
			m_ObjFile->WriteChksum();
		}
	}

	//EXTDEF PUBDEF - Done.

	while( String.ReadFile(fptr) && !Done)
	{
		String.Trim();
//		String.ToUpper();
		LineNo++;
		found=0;
		Read=GetNextToken(String,temp,0);
		if(!Read)
			continue;
		Label.Empty();
		flag=2;
		if((idx=GetPOT(temp))!=-1)    //Pseudo op
		{
			flag=1;
			found=1;
		}
		else if((idx=GetMOT(temp))!=-1) //Machine op
			found=1;
		if(!found)
		{
			if( CanBeLabel(temp) )
			{
				Label=temp;
				Label.SetLength( LabelLength );
			}
			Read+=GetNextToken(String,temp,Read);
			if((idx=GetPOT(temp))!=-1)    //Pseudo op
			{
				flag=1;
				found=1;
			}
			else if((idx=GetMOT(temp))!=-1) //Machine op
				found=1;
		}
		if(!found)//still no op code found-
		{
			Error("Invalid Mnemonic - "); //Should have been detected and reported in Pass 1 itself. Checked anyway!
			temp.Print();
			continue;
		}

		mne_op=temp;
		// printing assembly listing
		fprintf(fpList,"\n%3ld %0.4d %0.4X ",LineNo,LC, LC);

		if(flag==2)	//Machine op
		{
			ret=GetOperands();
			if( ret == 0 ) 		//no operands
			{
				while(idx!=-1)
				{
					pmot=Tables.GetMOTAt(idx);
					if( (pmot->type==0 && pmot->Accumulator==0
							&& pmot->SegmentRegister==0 && pmot->Reg_op==0 )
								|| pmot->type == 1)
						break;
					idx=GetNextMOT(temp,idx);
				}
				ret = CheckType(pmot,2);
				if (ret == -1 )
					break;
				LC++;
				if(pmot->type==1)
					LC++;
			}
			else //one or two operand
			{
				while(idx!=-1)
				{
					pmot=Tables.GetMOTAt(idx);
					ret=CheckType(pmot,2);
					if (ret == -1 )
						break;
					else if( ret !=0 )
					{
						LC += ret;
						break;
					}
					idx=GetNextMOT(temp,idx);
				}
				if( ret == -1 )
					continue;
				if(idx == -1 )
				{
					Error("Unable to handle mnemonic.");
					continue;
				}
			}
		}
		else //Pseudo op
			HandlePseudoOP(idx,2);

	}//end of while loop
	WriteSymbolTable(fpList);
	if(Errors)
		return 0;
	else
		return 1;
}

int CAssemble::GetMOT(CString & op_code)
{
	return GetNextMOT(op_code,-1);
}

int CAssemble::GetNextMOT(CString &op_code, int start)
{
	int i,max;
	MOT *temp;
	max=Tables.GetMOTCount();
	for(i=start+1;i<max;i++)
	{
		temp=Tables.GetMOTAt(i);
		if(op_code.Compare(temp->mne_op)==0)
			return i;
	}
	return -1;
}

int CAssemble::GetPOT(CString & op_code)
{
	int i,max;
	POT *temp;
	max=Tables.GetPOTCount();
	for(i=0;i<max;i++)
	{
		temp=Tables.GetPOTAt(i);
		if(op_code.Compare(temp->p_op)==0)
			return i;
	}
	return -1;
}

int CAssemble::EvaluateExpression(CString & exp, int &error)
{
	List * Item;
	CString temp;
	int x = 0;
	error = 0;
	GetNextToken(exp,temp,0);
	if( IsImmediate(temp))
		x = temp.ToInteger();
	else if( (Item = Sym.GetItem(temp) ) != NULL )
		x = Item->value;
	else
		error = 1;
	return x;
}

int CAssemble::IsDelimiter(char c)
{
	if(c==',' ||  c==0  || c==' ' || c=='\t' || c=='\n' || c==';' )
		return 1;
	return 0;
}

int CAssemble::GetNextToken(CString &Str,CString &Token, int start)
{
	int x,y,count=0;

	for(x=start; IsDelimiter(Str[x]) && Str[x] !=0  && Str[x] != ';' ;x++,count++); //Skipping Delimiters
	if(Str[x]==0 || Str[x]==';')
	{
		Token.Empty();
		return 0;
	}
	Token.Empty();
	for(; ! IsDelimiter(Str[x]) ;x++,y++,count++)
			Token += Str[x];
	return count;
}

// ######################  Pseudo op handlers ######################
int CAssemble::HandlePseudoOP(int op_id, int pass)
{
	int ret =1;
	switch(op_id)
	{
		case 0: //END
			if(pass==1)
				ret=END_1_pop();
			else
				ret=END_2_pop();
			break;
		case 1: //ASSUME
			if(pass==2)
				ret=ASSUME_2_pop();
			break;
		case 2: //DD
			ret = DataDefinition( pass, 4);
			break;
		case 3: //DB
			ret = DataDefinition( pass, 1);
			break;
		case 4: //DW
			ret = DataDefinition( pass, 2);
			break;
		case 5: //EQU
			if(pass==1)
				ret=EQU_1_pop();
			break;
		case 6: //SEGMENT
			if( pass == 1)
				ret=SEGMENT_1_pop();
			else
				ret=SEGMENT_2_pop();
			break;
		case 7: //ENDS
				ret=ENDS_pop();
			break;
		case 8: //EXTRN
			if( pass == 1 )
				ret=EXTRN_1_pop();
			break;
		case 9: //PUBLIC
			if( pass == 1 )
				ret = PUBLIC_1_pop();
			break;
		case 10: //GLOBAL
			if( pass == 1)
				ret = GLOBAL_1_pop();
			break;
		case 11: //PROC
			PROC_pop( pass );
			break;
		case 12: //ENDP
			ENDP_pop( );
			break;
		default:
			Error("Unhandled Pseudo op.\nPlease get an updated version of this Assembler");
			return 0;
	}
	if(ret && pass == 2  )
	{
		if( op_id < 2 || op_id > 4 ) //not DB DW DD
		{
			for(int i=0;i<7;i++)
				fprintf(fpList,"    ");
		}
		String.WriteFile(fpList);
	}
		return ret;
}

int CAssemble::ASSUME_2_pop()
{
	int Segment=-1, W, S, i;
	CString Str, Temp;

	Str = String.SubString(Read);

	for( i = 0 ; ; i++ )
	{
		if( Str[i]==',' || Str[i]==0 )
		{
			if(Segment == -1 )
			{
				Error("Illegal assume");
				return 0;
			}

			Temp.Trim();
			Temp.SetLength( LabelLength );
			switch (Segment)
			{
				case 0:
					ES=Temp; break;
				case 1:
					CS=Temp; break;
				case 2:
					SS=Temp; break;
				case 3:
					DS=Temp; break;
				case -1:
					Error("Illegal assume");
					return 0;
			}
			Segment=-1;
			Temp.Empty();
			if(Str[i]==0)
				break;
		}
		else if( Str[i]==':' )
		{
			Temp.Trim();
			Segment=Tables.IsRegister(Temp, W, S) ;
			if( Segment == -1 || !S )
			{
				Error("Illegal assume");
				return 0;
			}
			Temp.Empty();
		}
		else
			Temp += Str[i];

	}
	return 1;
}
/*
int CAssemble::DD_1_pop(int size)
{
	int i, sofar, count=0, ret;
	List Entry;
	CString Token;
	if(CurrentSegment.IsEmpty())
	{
		Error("Code or data is not within any segment");
		return -1;
	}

	//counting no of bytes to allocate
	sofar=Read;
	do
	{
		i = GetNextToken(String,Token,sofar);
		sofar += i;
		if (i)
		{
			count++;
		}
	}while (i);

	if(!Label.GetLength())
	{
		Warning("DD pseudo op with out name!");
		return 1;
	}
	if( !count )
	{
		Error("DD pseudo op without parameters");
		return 1;
	}
	Label.CopyTo(Entry.symbol);
	Entry.value = LC; //count * size;
	Entry.Relocation='R';
	Entry.CombineClass = ' ';
	Entry.Type='D';
	LC += count * size ;
	CurrentSegment.CopyTo(Entry.Segment);
	ret=Sym.AddItem(Entry);
	if(ret==-1)
	{
		Error("Symbol table full. Insufficient memory!");
		return 0;
	}
	return 1;
} */


int CAssemble::DataDefinition(int pass, int size)
{
	int max   = String.GetLength();
	int Quote = 0, Number = 0, ret;
	CString Temp;
	long ValueD;
	WORD ValueW;
	BYTE ValueB;
	int Count = 0;
	List Entry;

	if(CurrentSegment.IsEmpty())
	{
		Error("Code or data is not within any segment");
		return -1;
	}
	if(pass == 1)
	{
		if(!Label.GetLength())
		{
			mne_op.Print();
			Warning(" pseudo op with out label!");
			return 1;
		}
	}
	//Counting
	for( int i = Read; i< max; i++)
	{
		if( Quote )
		{
			if (Number)
			{
				mne_op.Print("%s pseudo op -");
				Error("parameters invalid");
				return 0;
			}
			if( String[i]== Quote )
			{
				Quote = 0;
				for( int j = 0 ; j < Temp.GetLength(); j++ )
				{
					Count ++;
					if(pass == 2)
					{
						if( size == 1 )
						{
							ValueB = Temp[j];
							fprintf(fpList," %0.2X ",ValueB);
							//LEDATA
							m_ObjFile->WriteCode( &ValueB, size );
						}
						else if (size == 2)
						{
							ValueW = Temp[j];
							fprintf(fpList," %0.4X ",ValueW);
							//LEDATA
							m_ObjFile->WriteCodeData( &ValueW, size );
						}
						else if (size == 4)
						{
							ValueD = Temp[j];
							fprintf(fpList," %0.8X ",ValueD);
							//LEDATA
							m_ObjFile->WriteCodeData( &ValueD, size );
						}
					}

				}
				Temp.Empty();
			}
			else
				Temp += String[i];
		}
		else if( Number )
		{
			if( String[i] == '\,' )
			{
				if( pass == 2 )
				{
					if( size == 1 )
					{
						ValueB = Temp.ToInteger() ;;
						fprintf(fpList," %0.2X ",ValueB);
						//LEDATA
						m_ObjFile->WriteCode( &ValueB, size );
					}
					else if (size == 2)
					{
						ValueW = Temp.ToInteger() ;
						fprintf(fpList," %0.4X ",ValueW);
						//LEDATA
						m_ObjFile->WriteCodeData( &ValueW, size );
					}
					else if (size == 4)
					{
						ValueD = Temp.ToInteger();
						fprintf(fpList," %0.8X ",ValueD);
						//LEDATA
						m_ObjFile->WriteCodeData( &ValueD, size );
					}

				}
				Count ++ ;
				Number = 0;
				Temp.Empty();
			}
			else
				Temp += String[i];
		}
		else
		{
			if( String[i]== '\'' || String[i] == '\"' )
				Quote = String[i];
			if( String[i] >= '0' && String[i]  <= '9' )
			{
				Number = 1;
				Temp += String[i];
			}
		}
	}

	if ( Number && Temp.GetLength() )
	{
		if( pass == 2 )
		{
			if( size == 1 )
			{
				ValueB = Temp.ToInteger() ;;
				fprintf(fpList," %0.2X ",ValueB);
				//LEDATA
				m_ObjFile->WriteCode( &ValueB, size );
			}
			else if (size == 2)
			{
				ValueW = Temp.ToInteger() ;;
				fprintf(fpList," %0.4X ",ValueW);
				//LEDATA
				m_ObjFile->WriteCodeData( &ValueW, size );
			}
			else if (size == 4)
			{
				ValueD = Temp.ToInteger();
				fprintf(fpList," %0.8X ",ValueD);
				//LEDATA
				m_ObjFile->WriteCodeData( &ValueD, size );
			}
		}
		Count ++ ;
	}
	if(Quote)
	{
		Warning("End quote not found; assumed.");
	}
	if(pass == 1)
	{
		if( !Count )
		{
			Error("Pseudo op without parameters");
			return 1;
		}
		Label.CopyTo(Entry.symbol);
		Entry.value = LC; //count * size;
		Entry.Relocation='R';
		Entry.CombineClass = ' ';
		CurrentSegment.CopyTo(Entry.Segment);
		if( size == 1 )
			Entry.Type='B';

		else if(size == 2 )
			Entry.Type='W';
		else if (size == 4 )
			Entry.Type = 'D';
		else
			Entry.Type = ' ';
		ret=Sym.AddItem(Entry);
		if(ret==-1)
		{
			Error("Symbol table full. Insufficient memory!");
			return 0;
		}
	}
	LC += ( Count * size );
	return 1;

}

int CAssemble::SEGMENT_1_pop()
{
	WORD static SegmentIndex = 0 ;
	List Entry;
	int ret;
	int More;
	CString Temp;
	LC=0;
	if(	! CurrentSegment.IsEmpty() )
	{
		Error("Nested Segments");
		return 0;
	}
	if(!Label.GetLength())
	{
		Warning("Segment pseudo op with out name!");
		return 1;
	}
	Label.SetLength( LabelLength );
	CurrentSegment=Label;
	Label.CopyTo(Entry.symbol);
	SegmentIndex ++;
	Entry.Index = SegmentIndex;
	Entry.GroupIndex = 0;
	Entry.CombineClass = ' ';
	Entry.Relocation=' ';
	Entry.Type='S';
	Entry.value = 0;
	Label.CopyTo(Entry.Segment);
	Entry.CombineClass = ' ';	//Private
	Entry.Align = 3;			//Para
	while( 1 )
	{
		More = GetNextToken( String, Temp, Read );
		if( ! More )
			break;
		Read += More;
		if( Temp.Compare("PUBLIC") == 0)
			Entry.CombineClass = 'P';
		else if( Temp.Compare("STACK") == 0)
			Entry.CombineClass = 'S';
		else if( Temp.Compare("BYTE") == 0)
			Entry.Align = 1;
		else if( Temp.Compare("WORD") == 0)
			Entry.Align = 2;
		else if( Temp.Compare("PARA") == 0)
			Entry.Align = 3;
		else if( Temp.Compare("PAGE") == 0)
			Entry.Align = 4;
		else
		{
			Error("\nExtra characters on line");
			return 0;
		}
/*		CombineClass;	// 	P-public, S-Stack ' 'Private -> Segment
		BYTE Align;		//  0-Absolute, 1-Byte, 2-Word, 3-Para, 4-Page,*/
	}
	ret=Sym.AddItem(Entry);
	if(ret==-1)
	{
		Error("Symbol table full. Insufficient memory!");
		return 0;
	}
	return 1;
}

int CAssemble::SEGMENT_2_pop()
{
	List *Item;
	LC=0;
	Label.SetLength( LabelLength );
	CurrentSegment=Label;
	Item = Sym.GetItem(Label);

	// Write LEDATA Header
	m_ObjFile->AcceptCodeHdr( Item->Index, 0); //Segment Index
	return 1;
}

int CAssemble::ENDS_pop()
{
	List *Item;
	if(CurrentSegment.IsEmpty())
	{
		Error("ENDS without corresponding SEGMENT directive.");
		return 0;
	}
	Item = Sym.GetItem(CurrentSegment);
	Item->SegmentLength = LC;
	CurrentSegment.Empty();
	return 1;
}
int CAssemble::GLOBAL_1_pop()
{
	return 1;
}

int CAssemble::PROC_pop( int pass )
{
	CString Temp;
	List Entry;
	int ret;

	CurrentProcedure = Label;
	ProcedureType = 'N';
	GetNextToken( String, Temp, Read );
	if( Temp.Compare("FAR") == 0 )
		ProcedureType = 'F';
	if( pass == 1 )
	{
		Label.CopyTo( Entry.symbol );
		Entry.value = LC;
		Entry.CombineClass = ' ';
		CurrentSegment.CopyTo( Entry.Segment );
		Entry.Relocation = 'R';
		Entry.Type = ProcedureType ;

		ret=Sym.AddItem(Entry);
		if(ret==-1)
		{
			Error(" Symbol table full. Insufficient memory!");
		}
	}
	return 1;
}

int CAssemble::ENDP_pop( )
{
	CurrentProcedure.Empty();
	ProcedureType = ' ';
	return 1;
}

int CAssemble::EXTRN_1_pop()
{
	List Entry;
	int ret;
	int max;
	CString Temp;
	if( !Label.IsEmpty() )
	{
		Error("Invalid label");
		return 0;
	}
	Operand1.Empty();
	Operand2.Empty();
	max = String.GetLength();
	for( int i = Read ; i < max ; i++ )
	{
		if( String[i] == ':' )
		{
			GetNextToken(String, Operand2, i+1);
			break;
		}
		Temp +=String[i];
	}
	Temp.Trim();
	Operand1=Temp;

	if( ! Operand2.GetLength() )
		Entry.Type = 'W'; //Word; default
	if( !Operand1.IsEmpty() )
	{
		CurrentSegment.CopyTo(Entry.Segment);
		Operand1.SetLength( LabelLength );
		Operand1.CopyTo(Entry.symbol);
		Entry.Relocation=' ';
		//  Type N-Near, F-Far, B-Byte, W-Word, D-Double A-Absolute
		if( Operand2.Compare("NEAR") == 0 )
		{
			Entry.Type='N';
		}
		else if( Operand2.Compare("FAR") == 0 )
		{
			Entry.Type='F';
		}
		else if( Operand2.Compare("BYTE") == 0 )
		{
			Entry.Type='B';
		}
		else if( Operand2.Compare("WORD") == 0 )
		{
			Entry.Type='W';
		}
		else if( Operand2.Compare("DOUBLE") == 0 )
		{
			Entry.Type='D';
		}
		else if( Operand2.Compare("ABS") == 0 )
		{
			Entry.Type='A';
		}
		Entry.CombineClass = 'E'; //External data
		Entry.value=0;
		ret=Sym.AddItem(Entry);
		if(ret==-1)
		{
			Error("Symbol table full. Insufficient memory!");
			return 0;
		}
	}
	return 1;
}

int CAssemble::PUBLIC_1_pop()
{
	CString Temp;
	int i, count=0;
	List Entry;

	while ( 1 )
	{
		i = GetNextToken( String, Temp, Read );
		if( !i )
			break;
		count++;
		Read += i;
		Temp.SetLength( LabelLength );
		Temp.CopyTo(Entry.symbol);
		Sym.MakePublic(Temp);
	}
	if( ! count )
	{
		Error("Public pseudo op without parameter");
		return 0;
	}
	return 1;
}

int CAssemble::EQU_1_pop()
{

	List Entry;     //variable to hold details before it is put in CSymLit
	int ret=1;
	CString Temp;

	if(Label[0]!=0)
	{
		Label.CopyTo(Entry.symbol);
		CurrentSegment.CopyTo(Entry.Segment);
		Temp = String.SubString(Read);
		Temp.Trim();
		Entry.CombineClass = ' ';
		if(Temp[0]=='$' && IsDelimiter(Temp[1]) )
		{
			Entry.value = LC;
			Entry.Relocation='R';
		}
		else
		{
			Entry.value=EvaluateExpression( Temp, ret);
			Entry.Relocation='A';
		}
		Entry.Type='A';
		ret=Sym.AddItem(Entry);
		if(ret==-1)
		{
			Error("Symbol table full!\nCannot proceed!");
			ret=0;
		}
		else if(ret==-2)
		{
			Warning(" defined more than once.");
			printf(" - %s",Entry.symbol);
		}
	}
	else
	{
		Warning("EQU pseudo op without Label!");
	}
	return ret;
}

int CAssemble::END_1_pop()
{
	if(!CurrentSegment.IsEmpty())
	{
		Warning("ENDS pseudo op not specified; assumed.");
	}

	Done=1;
	return 1;
}

int CAssemble::END_2_pop()
{
	Done=1;
	return 1;
}

int CAssemble::GetOperands()
{
	long x,y,sofar;
	int found=0, ptr, begining=1;
	CString Temp;
	Operand1.Empty();
	Operand2.Empty();
	Ptr=0;		//Not set
	sofar=Read;
	while(1)
	{
		x=GetNextToken(String,Temp,sofar);
		if(!x)
			break;
		sofar += x;
		if(begining && Temp.Compare("SHORT")==0 )
		{
			Ptr=1;
			continue;
		}
		else if(begining && (Temp.Compare("WORD")==0 || Temp.Compare("FAR")==0 ||
				 Temp.Compare("BYTE")==0 || Temp.Compare("DWORD")==0 ) )
		{
			if(Temp[0] == 'W'|| Temp[0] == 'w' )
				ptr=2;
			else if(Temp[0] == 'B' || Temp[0] == 'b')
				ptr= 1;
			else
				ptr=4;
			x=GetNextToken(String,Temp,sofar);
			if(!x)
			{
				Error("Invalid operands");
				break;
			}
			sofar +=x;
			if(Temp.Compare("PTR") == 0)
			{
				Ptr=ptr;
				continue;
			}
			else
			{
				Error("Invalid operands");
				break;
			}

		}
		begining=0;
		if(!found)
		{
			if(Operand1.GetLength())
				Operand1 += " ";
			Operand1 += Temp;
		}
		else
		{
			if(Operand2.GetLength())
				Operand2 += " ";
			Operand2 += Temp;
		}
		for(y=sofar;String[y]!=',' && String[y]!=0 && String[y]!=';' &&
					IsDelimiter(String[y]) ; y++);
		if( String[y] == ',' )
		{
			if(!found)
			{
				found ++;
				begining=1;
			}
			else
			{
				Error("Extra characters in the line.");
				break;
			}
		}
	}
	if(Operand2.GetLength())
		return 2;
	else if(Operand1.GetLength())
		return 1;
	else
		return 0;
}

int CAssemble::IsImmediate(CString & Op)
{
	CString Operand(Op);
	Operand.ToUpper();
	int i, Hex=0, Bin=0, Oct=0; //, Dec=0;
	int ret = 1;
	if(Operand.IsEmpty())
		return 0;
	if(Tables.IsRegister(Operand) != -1 )
		return 0;
	switch(Operand[Operand.GetLength() -1 ] )
	{
		case 'H':
			Hex = 1; break;
		case 'O':
			Oct = 1; break;
		case 'B':
			Bin = 1; break;
	}

	for (i=0;Operand[i]!=0;i++)
	{
		if(Operand[i+1]==0 && ( Operand[i] == 'D' ||
				Operand[i] == 'O' || Operand[i] == 'B' ||
					Operand[i] == 'H' ) )
			break;

		if( (Operand[i] =='0' || Operand[i] =='1') );
		else if( Operand[i] >='2' && Operand[i] <='7')
		{
			if(Bin)
			{
				ret = 0;
				break;
			}
		}
		else if( Operand[i] >='8' && Operand[i] <='9')
		{
			if(Bin || Oct)
			{
				ret = 0;
				break;
			}
		}
		else if( (Operand[i] >='A' && Operand[i]<='F') || (Operand[i] >='a' && Operand[i]<='f') )
		{
			if(!Hex)
			{
				ret = 0;
				break;
			}
		}
		else
		{
			ret = 0 ;
			break;
		}
	}

	return ret;

}

int CAssemble::CanBeLabel(CString & Op)  //CouldBeLabel
{
	CString Operand = Op;
	Operand.ToUpper();
	int i=0, max;
	max=Operand.GetLength();
	if(! (Operand[i] >= 'A' && Operand[i] <= 'Z' ) )
		return 0;

	for(; i<max ; i++ )
	{
		if(!(Operand[i] >= 'A' && Operand[i] <= 'Z' ) &&
			!(Operand[i]>='0' && Operand[i]<='9') &&
			Operand[i]!='_')
			return 0;
	}
	if( max == 2 && Tables.IsRegister(Operand) != -1 )
		return 0;

	return 1;
}

int CAssemble::IsAddress(CString & Operand)  //CouldBeAddress
{
	char c;
	c = Operand[0];
	if( c == '[' || c == '(')//address
		return 1;
	else if( CanBeLabel(Operand) )
		return 1;
	else
		return 0;
}

int CAssemble::IsDirectAddress(CString & Operand)  //CouldBeAddress
{
	char c;
	CString Temp;
	c = Operand[0];
	if( c == '[' || c == '(')//address
	{
		Temp = StripBraces(Operand);
//		if( CanBeLabel(Temp) || IsImmediate(Temp) )
		if( IsImmediate(Temp) )
			return 1;
	}
	else if( CanBeLabel(Operand) )
		return 1;
	return 0;
}

CString CAssemble::StripBraces(CString & Operand)
{
	CString Target;
	int max,i;
	Target.Empty();
	max=Operand.GetLength();
	if(max<2)
		return CString ();
	if( ( Operand[0]!='(' && Operand[0]!='[' ) ||
			( Operand[max-1]!=')' && Operand[max-1]!=']') );
	else
	{
		for(i=1;i<max-1;i++)
			Target += Operand[i];
	}
	return CString(Target);
}

int CAssemble::IsNear(CString & Operand)
{
//Incomplete implementation*******************************8
	if( Operand.GetLength() )
		return 1;
	else
		return 0;
}

int CAssemble::CheckType(MOT *pmot, int pass)
{
	int ret=0;
	if(CurrentSegment.IsEmpty())
	{
		Error("Code or data is not within any segment");
		return -1;
	}
	switch(pmot->type)
	{
		case 0 :
			ret= CheckType0(pmot, pass);
			break;
		case 1 :
			ret = CheckType1(pmot, pass);
			break;
		case 2 :
			ret= CheckType2(pmot, pass);
			break;
		case 3 :
			ret= CheckType3(pmot, pass);
			break;
		case 4 :
			ret= CheckType4(pmot, pass);
			break;
		case 5 :
			ret= CheckType5(pmot, pass);
			break;
		case 6 :
			ret= CheckType6(pmot, pass);
			break;
		case 7 :
			ret= CheckType7(pmot, pass);
			break;
		case 8 :
			ret= CheckType8(pmot, pass);
			break;
		default:
			ret = 0;
	}
	if( pass == 2 && mne_op.Compare("RET") == 0 && ProcedureType == 'F') //Special handler
	{
		MachineCode[0] |= 8;
	}
	//Generating List file and Object code
	if(ret && pass == 2)
	{
		fprintf(fpList,"T%d",pmot->type);
		for(int i=0; i< ret; i++)
			fprintf(fpList," %0.2X ",MachineCode[i]);
		for( ; i< 7; i++)
			fprintf(fpList, "    ");
		String.WriteFile(fpList);
		m_ObjFile->WriteCode( MachineCode, ret );
	}
	return ret;
}

int CAssemble::CheckType0(MOT *pmot,int pass)
{
	BYTE Reg;
	int val;
	if(!Operand2.IsEmpty())
		return 0;
	if( pass == 2)
	{
		MachineCode[0]=pmot->op_code;
	}

	if(Operand1.IsEmpty())
	{
		if( pmot->Accumulator !=0
			|| pmot->SegmentRegister != 0 || pmot->Reg_op != 0 )
			return 0;
		else
			return 1;
	}
	else
	{
		val = Tables.IsRegister(Operand1);
		if( val == -1 )
			return 0;
		Reg = val;
		if( pmot->Accumulator ==0
			&& pmot->SegmentRegister == 0 && pmot->Reg_op == 0 )
			return 0;

		if( pmot -> Accumulator == 0 && Reg==0 )
			return 1;
		if( pmot->Reg_op && pass== 2 )
		{
			Reg = Reg << ( pmot->Reg_op - 2) ;
			MachineCode[0] |= Reg;
			return 1;
		}
	}

	return  1 ;
}

int CAssemble::CheckType1(MOT *pmot,int pass)
{
	if( pmot->Accumulator !=0
			|| pmot->SegmentRegister != 0 || pmot->Reg_op != 0 )
		return 0;
	if(pass == 2)
	{
		MachineCode[0]=pmot->op_code;
		MachineCode[1]=pmot->ext_op_code;
	}
	return 2;
}

int CAssemble::CheckType2(MOT *pmot,int pass)
{
	//2. Immediate data [ 1 or 2 bytes ] No Segment Register
	List *Item;
	MOT *pmot1;
	CString tmp;
	int Wide1, Seg1, Wide2, Seg2, Reg1, Reg2;
	int ret=0, i, j, disp, ret1, Label1=0;

	if(Ptr)
		return 0;
	Reg1 = Tables.IsRegister(Operand1,Wide1,Seg1);
	if( Reg1!=-1 && Seg1 != 0 ) //Segment Register
		return 0;
	if(Operand2.GetLength()==0) //Only one operand (eg: RET instruction)
	{
		if( pmot->Accumulator || pmot->Reg_op )
		{
			return 0;
		}
		if(IsImmediate(Operand1))
		{
			if(pass==2)
			{
				disp=Operand1.ToInteger();
				MachineCode[1]=GetLow(disp);
				MachineCode[2]=GetHigh(disp);
			}
			ret  = 3;
		}
		else if( CanBeLabel(Operand1) )//Not an address
		{
			if(pass==2)
			{
				Item=Sym.GetItem(Operand1) ;
				if(!Item)
				{
					Error("Label not defined ");
					Operand1.Print("- %s");
					return -1;
				}
				if( Item->Relocation != 'A')
				{
					Error("Invalid operand");
					Operand1.Print("- %s");
					return -1;
				}

				disp=Item->value;
				MachineCode[1]=GetLow(disp);
				MachineCode[2]=GetHigh(disp);
			}
			ret= 3;
		}
		if(pass == 2)
			MachineCode[0]=pmot->op_code;
		return ret;
	}
	//two operands
	//Destination has to be register
	//Source has to be immediate or Segment label

	Reg2 = Tables.IsRegister(Operand2,Wide2,Seg2);
	if( Reg2!=-1 || Reg1 ==-1) //Source is register or Destination is not
		return 0;
	else if(IsImmediate(Operand1) )//Destination immediate
	{
		Error("Destination cannot be immediate");
		return 0;
	}
	else if ( CanBeLabel(Operand1) )
		return 0;					//Destination not Register

	if( (Reg1 == 0 && pmot->Accumulator) || pmot->Reg_op)
	{
		ret = Wide1 + 2;
		if(pass == 2)
		{
			MachineCode[0]=pmot->op_code;
			if(pmot->Reg_op)
			{
				Reg1 = Reg1 << ( pmot->Reg_op -2 );
				MachineCode[0] |= Reg1;
				if(Wide1)
				{
					MachineCode[0] |= (Wide1 << pmot->Reg_op + 1);
				}
			}
			else if(pmot->w)
			{
				MachineCode[0] |= Wide1;
			}
		}

		if( IsImmediate(Operand2) )
		{
			if(pass==2)
			{
				disp=Operand2.ToInteger();
				if(disp >=256 && !Wide1)
				{
					Error("Operands incompatible ");
					return -1;
				}
				if(Wide1)
				{
					MachineCode[1]=GetLow(disp);
					MachineCode[2]=GetHigh(disp);
				}
				else
					MachineCode[1]=GetLow(disp);
			}
		}
		else if( (Label1=CanBeLabel(Operand2))==1 )
		{
			if(pass==2)
			{
				Item = Sym.GetItem(Operand2);
				if(Item == NULL)
				{
					Error("Label undefined");
					Operand2.Print("- %s");
					return -1;
				}
				if( Item->Relocation == 'R' ) 	// Label is address
				{
					return 0;
				}
				if( Item->Type == 'S' ) 	// Label is Segment
				{
					//Writing Fixup Record
					BYTE Fixdat = 0x54;
					WORD Locat = m_ObjFile->GetRelativeOffset() + 1;
					BYTE High, Low;
					Locat |= 0xC800;
					High = GetHigh( Locat );
					Low  = GetLow ( Locat );
					m_ObjFile->WriteFixupRec( & High,   1 );
					m_ObjFile->WriteFixupRec( & Low,    1 );
					m_ObjFile->WriteFixupRec( & Fixdat, 1 );
					//Target index
					m_ObjFile->WriteFixupIndex( Item -> Index );
				}

				disp = Item->value;
				if(disp >=256 && !Wide1)
				{
					Error("Operands incompatible ");
					return -1;
				}
				if(Wide1)
				{
					MachineCode[1]=GetLow(disp);
					MachineCode[2]=GetHigh(disp);
				}
				else
					MachineCode[1]=GetLow(disp);
			}
		}
		else
			ret = 0;

	}
	if(ret && Label1) // Allocating more bytes if necessary
	{
		i=idx;
		while( 1 )
		{
			tmp=pmot->mne_op;
			i=GetNextMOT(tmp,i);
			if( i == -1 )
				break;
			pmot1=Tables.GetMOTAt(i);
				if( (ret1=CheckType(pmot1 ,1) ) > 0)
			{
			   if( ret1 >= ret )
			   {
					if(pass==2)
					{
						for(j=ret;j<ret1;j++)
							MachineCode[j]=0x90;
					}
					ret = ret1;
				}
				break;
			}
		}
	}
	return ret;
}

int CAssemble::CheckType3(MOT *pmot,int pass)
{
	int val;
	List *Item;
	if(!Operand2.IsEmpty())
		return 0;
	val=IsImmediate(Operand1);
	if( !val && IsAddress( Operand1 ) )
		return 0;
	else if(pass == 2)
	{
		MachineCode[0]=pmot->op_code;
		if(val)//Immediate
		{
			val=Operand1.ToInteger();
			if(val >=256 )
			{
				Error("Operand cannot be greater than 255 ");
				return -1;
			}
			MachineCode[1]=val;
		}
		else//Label
		{
			Item=Sym.GetItem(Operand1);
			if(Item == NULL)
			{
				Error("Label undefined - ");
				Operand1.Print("- %s");
				return -1;
			}
			if( Item->Relocation != 'A')
			{
				Error("Invalid Operand");
				return -1;
			}
			val=Item->value;
			if(val >=256 )
			{
				Error("Operand cannot be greater than 255 ");
				return -1;
			}
			MachineCode[1]=val;
		}
	}
	return  2 ;
}

int CAssemble::CheckType4(MOT *pmot,int pass)
{
	List * Item;
	BYTE Type = 'A';
	int ret = 0, wide1, seg1, wide2, seg2, disp=0, Reg1, Reg2, error;
	BYTE D=0, W=0;
	if(Ptr==1)      // Short Type 5
		return 0;
	else if(Ptr==4) // Far address Type 6
		return 0;
	if(!Operand2.GetLength())//only one operand
	{
		if(IsDirectAddress(Operand1) && pmot->Accumulator == 0)
			ret = 3;
		if(pass==2)
		{
			MachineCode[0]=pmot->op_code;
			disp=GetAddress(Operand1,error);
			if(error)
				return -1;
			disp = disp - (LC + 3); //Difference
			Type = 'D';
			Item = Sym.GetItem( Operand1 );
			MachineCode[1]=GetLow(disp);
			MachineCode[2]=GetHigh(disp);
		}
	}
	else//Two operands one must be an accumulator
	{
		if(!pmot->Accumulator );
		else
		{
			Reg1 = Tables.IsRegister(Operand1, wide1, seg1);
			Reg2 = Tables.IsRegister(Operand2, wide2, seg2);
			if( ( Reg1==0 && IsDirectAddress(Operand2) ) ||
				( Reg2==0 && IsDirectAddress(Operand1) ) )
				ret = 3;
			if(pass == 2 && ret)
			{
				if(Reg1 == 0)
				{
					if(wide1)
						W=1;
					disp=GetAddress(Operand2,error);
					if(error)
						return -1;
					Item = Sym.GetItem(Operand2);
				}
				else//Reg2=0 Accumulator is source
				{
					if(wide2)
						W=1;
					D=1;
					disp=GetAddress(Operand1,error);
					if(error)
						return -1;
					Item = Sym.GetItem(Operand1);
				}

				//Checking if fixup is required

				MachineCode[0]= pmot->op_code ;
				if(pmot->w)
					MachineCode[0] |= W;
				if(pmot->d)
					MachineCode[0] |= D;
				MachineCode[1]=GetLow(disp);
				MachineCode[2]=GetHigh(disp);
			}//end of if (pass==2)
		}
	}
	if(ret && Item != NULL && pass == 2 &&
			( ( Type =='D' && Item->CombineClass == 'E' ) || Type == 'A') )
	{
		//Writing Fixup Record
		CString Seg = Item->Segment;
		BYTE Fixdat = 0x50;
		WORD TargetIndex;
		if( Item->CombineClass == 'E' )
		{
			TargetIndex = Item->Index;
			Fixdat |= 6;
		}
		else
		{
			List *Item1 = Sym.GetItem( CString( Item->Segment ) );
			TargetIndex = Item1->Index;
			Fixdat |= 4;
		}
		WORD Locat = m_ObjFile->GetRelativeOffset() + 1;
		BYTE High, Low;
		if( Type == 'D' )
			Locat |= 0x8400;
		else
			Locat |= 0xC400;
		High = GetHigh( Locat );
		Low  = GetLow ( Locat );
		m_ObjFile->WriteFixupRec( & High,   1 );
		m_ObjFile->WriteFixupRec( & Low,    1 );
		m_ObjFile->WriteFixupRec( & Fixdat, 1 );
		//Target index
		m_ObjFile->WriteFixupIndex( TargetIndex );
		//Fixup end
	}
	return  ret ;
}

int CAssemble::CheckType5(MOT *pmot,int pass)
{
	int ret = 0, error, disp;
	BYTE dispb;
	if(Ptr > 1) //Not Short
		return 0;
	if(!Operand2.GetLength())//only one operand
	{
		if(IsDirectAddress(Operand1) && pmot->Accumulator == 0)
			ret = 2;
		if(pass==2 && ret )
		{
			MachineCode[0]=pmot->op_code;
			disp=GetAddress(Operand1,error);
			if(error)
				return -1;

			//Getting displacement
			disp = disp - (LC + 2);
			if(disp > 127 || disp < -128)
			{
				Error("Operand too large");
				return -1;
			}
			dispb= (BYTE) disp;
			MachineCode[1]=dispb;
			//Checking if fixup is required
			List * Item = Sym.GetItem(Operand1);
			if( !IsImmediate(Operand1) && Item->CombineClass == 'E')
			{

				//Writing Fixup Record
				CString Seg = Item->Segment;
				BYTE Fixdat = 0x50;
				WORD TargetIndex;
				TargetIndex = Item->Index;
				Fixdat |= 6;
				WORD Locat = m_ObjFile->GetRelativeOffset() + 1;
				BYTE High, Low;
				Locat |= 0x8000;
				High = GetHigh( Locat );
				Low  = GetLow ( Locat );
				m_ObjFile->WriteFixupRec( & High,   1 );
				m_ObjFile->WriteFixupRec( & Low,    1 );
				m_ObjFile->WriteFixupRec( & Fixdat, 1 );
				//Target index
				m_ObjFile->WriteFixupIndex( TargetIndex );
				//Fixup end
			}

		}

	}
	return   ret;

}

int CAssemble::CheckType6(MOT *pmot,int pass)
{
	List *Item;
	//Determine if call is far - 'FAR PTR' or label is declared as far
	if(Ptr>0 && Ptr<4)
		return 0;
	if( !Operand2.IsEmpty() ) // only one operand possible
		return 0;
	if( !IsDirectAddress( Operand1 ) )
		return 0;
	else if(pass == 2 )
	{
		//finding address
		Item = Sym.GetItem(Operand1);
		if(Item==NULL)
		{
			Error("Label undefined -");
			Operand1.Print();
			return -1;
		}
		MachineCode[0]=pmot->op_code;
		MachineCode[1]=GetLow(Item->value);
		MachineCode[2]=GetHigh(Item->value);
		MachineCode[3]=0;
		MachineCode[4]=0;

		//Fixup information
		BYTE Fixdat = 0x50;
		WORD TargetIndex;

		if( Item->CombineClass == 'E' )
		{
			TargetIndex = Item->Index;
			Fixdat |= 6;
		}
		else
		{
			List *Item1 = Sym.GetItem( CString( Item->Segment ) );
			TargetIndex = Item1->Index;
			Fixdat |= 4;
		}
		WORD Locat = m_ObjFile->GetRelativeOffset() + 1;
		BYTE High, Low;
		Locat |= 0xCC00;
		High = GetHigh( Locat );
		Low  = GetLow ( Locat );
		m_ObjFile->WriteFixupRec( & High,   1 );
		m_ObjFile->WriteFixupRec( & Low,    1 );
		m_ObjFile->WriteFixupRec( & Fixdat, 1 );
		//Target index
		m_ObjFile->WriteFixupIndex( TargetIndex);
		//Fixup end
	}
	return  5;

}

int CAssemble::CheckType7(MOT *pmot,int pass)
{
	int ret=0;
	int disp=0;
	int i, times, error;
	int Reg=0, Mem = 0, No_of_Parts=0;
	int Wide,Seg, R;
	CString Operand, Part[3], Reg1, Reg2;
	int Index = 0, FixupRequired = 0;

	List *Item;

	BYTE RegOp=0, RegOp1, RegOp2, MemOp=0, D=0, W=0, S=0, Temp, NOP=0;
	if( (pmot->FarCall && Ptr!=4) || (Ptr==4 && !pmot->FarCall) )
		return 0;
	if( !Operand2.IsEmpty() && pmot->Reg_operand_valid )//Only one operand
		return 0;
	else
		RegOp = pmot->Reg_operand << 3 ;
	if(pass == 2)
		MachineCode[2]=0;
	for(times=1;times<=2;times++)
	{
		if(times==2 && Operand2.IsEmpty() )
			break;
		Part[0].Empty();
		Part[1].Empty();
		Part[2].Empty();
		if(times==1)
			Operand=Operand1;
		else
			Operand=Operand2;
		i=0;
		if(Operand[0]=='[' || Operand[0]=='(' )
		{
			Part[0] += Operand[0];
			i++;
		}
		for(No_of_Parts=0;Operand[i]!=0;i++)//Splitting into parts
		{
			if( Operand[i] == '[' || Operand[i] =='(' )
			{
				Part[No_of_Parts].Trim();
				No_of_Parts++;
			}
			if(No_of_Parts == 3)
				return 0;
			Part[No_of_Parts] += Operand[i];
		}
		//Processing Parts
		if( ( R=Tables.IsRegister(Part[0],Wide,Seg) ) != -1 )
		{
			if(Seg && !pmot->SegmentRegister)
				return 0;
			if(Seg )
			{
				if(S)//Two segment registers
				{
					Error("Two segment registers cannot be specified");
					return -1;
				}
				S=times;
			}

			if(Reg==2 && ( (Wide && !W) || (!Wide && W) ) )
			{
				Error("Operands are incompatible");
				return -1;
			}
			if(Wide)
				W=1;
			if(No_of_Parts == 0)
			{
				Reg++;
				if(Reg==1)
					RegOp1=R;
				else
					RegOp2=R;
				continue;
			}
			return 0;
		}
		else if( IsImmediate(Part[0]) )
		{
			if(No_of_Parts==0)
				return 0;				//Immediate data alone unacceptable
			disp=Part[0].ToInteger();
			if( disp >= 256)
			{
				if(pass==2)
				{
					MachineCode[2]=GetLow(disp);
					MachineCode[3]=GetHigh(disp);
				}
				disp =2;
			}
			else
			{
				if(pass==2)
					MachineCode[2] = (BYTE) disp;
				disp = 1;
			}
		}
		else if( CanBeLabel(Part[0]) )
		{
			disp=2;
			NOP=0;
			if(pass==2)
			{
				Item=Sym.GetItem(Part[0]);

				if(Item)
				{
					if(No_of_Parts==0)
						MemOp = 6; 	//	00---110 Mod/Reg/RM
					disp=Item->value;
					if( disp >= 256)
					{
						MachineCode[2]=GetLow(disp);
						MachineCode[3]=GetHigh(disp);
						disp=2;
					}
					else
					{
						MachineCode[2] = (BYTE) disp;
						if(No_of_Parts==0)
						{
							MachineCode[3]=0;
							disp=2;
						}
						else
						{
							MachineCode[3]=0x90;
							NOP=1;
							disp = 1;
						}
					}

					// Used by fixup
					if( Item->Relocation == 'R' )
					{
						List *Item1 = Sym.GetItem( CString( Item->Segment ) );
						Index = Item1->Index;
						FixupRequired = 1;
					}
					else if( Item->CombineClass == 'E')
					{
						Index = Item->Index;
						FixupRequired = 1;
					}
				}
				else
				{
					Error("Symbol not defined  - ");
					Part[0].Print();
					return -1;
				}

			}//end of if Pass==2
		}
		else if( IsImmediate( StripBraces(Part[0]) ) )
		{
			if(No_of_Parts != 0)
				return 0;
			Part[0]=StripBraces(Part[0]);
			disp=Part[0].ToInteger();
			if( disp >= 256)
			{
				if(pass==2)
				{
					MachineCode[2]=GetLow(disp);
					MachineCode[3]=GetHigh(disp);
				}
				disp =2;
			}
			else
			{
				if(pass==2)
					MachineCode[2] = (BYTE) disp;
				disp = 1;
			}
			MemOp = 6; 	//	00---110 Mod/Reg/RM
		}
		else
		{
			if (No_of_Parts==2)
				return 0;
			No_of_Parts++;
			Part[2]=Part[1];
			Part[1]=Part[0];
		}
		Mem++;
		if(times == 2)
			D=2; //Direction bit 1
		if( Part[1].GetLength() )
		{
			MemOp = CheckIndices(Part[1], Part[2], disp, pass, error);
			if ( error )
				return -1;
		}
	}//end of for times

	if( Reg >= 1 || Mem == 1  )
		ret = 2 + disp + NOP;
	if( pass==2 && ret )
	{
		if(S==2 && Reg==2) //Segment Register should not be represented as Mod/RM
		{
			Temp=RegOp1;
			RegOp1 = RegOp2;
			RegOp2=Temp;
		}
		if(Reg==2 || ( Mem==1 && Reg==1 ) )
			RegOp = RegOp1 << 3 ;
		if(Reg==2)
		{
			MemOp= RegOp2 | 192 ;	//to be represented as mod/rm
			if(S!=2)
				D=2;
		}
		else if (Reg + Mem == 1) //Only one operand
		{
			if(Reg)	//Only Register
				MemOp = RegOp1 | 192 ;
			else
				MemOp |= RegOp;
		}
		//  Fixup information
		//Writing Fixup Record
		if( FixupRequired )
		{
			BYTE Fixdat;
			if( Item -> CombineClass == 'E' && Item->Segment[0] != 0 )
				Fixdat = 0x06;
			else
				Fixdat = 0x54;
			WORD Locat = m_ObjFile->GetRelativeOffset() + 2;
			BYTE High, Low;
			Locat |= 0xC400;
			High = GetHigh( Locat );
			Low  = GetLow ( Locat );
			m_ObjFile->WriteFixupRec( & High,   1 );
			m_ObjFile->WriteFixupRec( & Low,    1 );
			m_ObjFile->WriteFixupRec( & Fixdat, 1 );
			//Frame index
			if( Item -> CombineClass == 'E' && Item->Segment[0] != 0 )
			{
				List *Item1 = Sym.GetItem( CString(Item->Segment) );
				m_ObjFile->WriteFixupIndex( Item1->Index );
			}
			//Target index
			m_ObjFile->WriteFixupIndex( Index );
			//Fixup end
		}

		MachineCode[0]=pmot->op_code;
		if(pmot->w)
			MachineCode[0]  |= W;
		if(pmot->d)
			MachineCode[0]  |= D;
		MachineCode[1]= RegOp | MemOp;
	}
	return ret; //ret;
}

int CAssemble::CheckType8(MOT *pmot,int pass)
{
	int ret=0, error;
	int disp=0;
	int i;
	int Reg=0, Mem = 0, No_of_Parts=0;
	int Wide, Seg, R, ImmOp;
	int Index = 0;
	int FixupRequired = 0;
	CString Part[3], Reg1, Reg2;
	BYTE MemOp=0, W=0, NOP;
	List *Item;

	if( Operand2.IsEmpty() )//two operands required
		return 0;
	//Processing the second operand (immediate operand
	if( !IsImmediate(Operand2) )
		return 0;
	if(pass ==2)
		ImmOp = Operand2.ToInteger();
	//Processing the first operand (Register / Memory)
	//Splitting into parts
	i=0;
	if(Operand1[0]=='[' || Operand1[0]=='(' )
	{
		Part[0] += Operand1[0];
		i++;
	}
	for(No_of_Parts=0;Operand1[i]!=0;i++)
	{
		if( Operand1[i] == '[' || Operand1[i] =='(' )
		{
			Part[No_of_Parts].Trim();
			No_of_Parts++;
		}
		if(No_of_Parts == 3)
			return 0;
		Part[No_of_Parts] += Operand1[i];
	}
	//Processing Parts
	R=Tables.IsRegister(Part[0],Wide,Seg);
	if( R != -1 ) //Register
	{
		MemOp= R | 192 ;	//to be represented as mod/rm
		if(Seg )
		{
			Error("Segment register cannot be specified");
			return 0;
		}
		if(Wide)
		{
			W=1;
			if(Ptr==1 || Ptr==4)//BYTE || DWORD
			{
				Error("Mismatch in operand sizes");
				return -1;
			}
		}
		if(No_of_Parts == 0)
			Reg = 1;
		else
			return 0;
	}
	else if( IsImmediate(Part[0]) )
	{
		if(No_of_Parts==0)         //Immediate data as destination
		{
			Error("Illegal immediate, cannot appear as destination");
			return -1;
		}
		disp=Operand1.ToInteger();
		if( disp >= 256)
		{
			if(pass==2)
			{
				MachineCode[2]=GetLow(disp);
				MachineCode[3]=GetHigh(disp);
			}
			disp =2;
		}
		else
		{
			if(pass==2)
				MachineCode[2] = (BYTE) disp;
			disp = 1;
		}
	}
	else if( CanBeLabel(Part[0]) )
	{
		disp=2;
		NOP=0;
		if(pass==2)
		{
			Item=Sym.GetItem(Part[0]);
			if(Item)
			{
				if(No_of_Parts==0)
					MemOp = 6; 	//	00---110 Mod/Reg/RM
				disp=Item->value;
				if( disp >= 256)
				{
					MachineCode[2]=GetLow(disp);
					MachineCode[3]=GetHigh(disp);
					disp=2;
				}
				else
				{
					MachineCode[2] = (BYTE) disp;
					if(No_of_Parts==0)
					{
						MachineCode[3]=0;
						disp=2;
					}
					else
					{
						NOP=1;
						disp = 1;
					}
				}
				//Fixup info
				if( Item->Relocation == 'R' )
				{
					List *Item1 = Sym.GetItem( CString( Item->Segment ) );
					Index = Item1->Index;
					FixupRequired = 1;
				}
				else if( Item->CombineClass == 'E')
				{
					Index = Item->Index;
					FixupRequired = 1;
				}
			}
			else
			{
				Error("Symbol not defined - ");
				Part[0].Print();
				return -1;
			}
		}//end of if Pass==2
	}
	else if( IsImmediate( StripBraces(Part[0]) ) )
	{
		if(No_of_Parts != 0)
			return 0;
		Part[0]=StripBraces(Part[0]);
		disp=Part[0].ToInteger();
		if( disp >= 256)
		{
			if(pass==2)
			{
				MachineCode[2]=GetLow(disp);
				MachineCode[3]=GetHigh(disp);
			}
			disp =2;
		}
		else
		{
			if(pass==2)
				MachineCode[2] = (BYTE) disp;
			disp = 1;
		}
		MemOp = 6; 	//	00---110 Mod/Reg/RM
	}
	else
	{
		if (No_of_Parts==2)
			return 0;
		No_of_Parts++;
		Part[2]=Part[1];
		Part[1]=Part[0];
	}
	if(!Reg)
	{
		Mem++;
		if(Ptr==2)
			W=1;
		else if(Ptr==4)
		{
			Error("Operand size override invalid");
		}
		if( Part[1].GetLength() )
		{
			MemOp = CheckIndices(Part[1], Part[2], disp, pass, error);
			if ( error )
				return -1;
		}
	}//end of if(!Reg)
	if(Reg  || Mem )
	{
		ret = 3 + disp + W + NOP;
	}
	if( pass==2 && ret )
	{
		if( FixupRequired )
		{
			BYTE Fixdat;
			if( Item -> CombineClass == 'E' && Item->Segment[0] != 0 )
				Fixdat = 0x06;
			else
				Fixdat = 0x54;
			WORD Locat = m_ObjFile->GetRelativeOffset() + 2;
			BYTE High, Low;
			Locat |= 0xC400;
			High = GetHigh( Locat );
			Low  = GetLow ( Locat );
			m_ObjFile->WriteFixupRec( & High,   1 );
			m_ObjFile->WriteFixupRec( & Low,    1 );
			m_ObjFile->WriteFixupRec( & Fixdat, 1 );
			//Frame index
			if( Item -> CombineClass == 'E' && Item->Segment[0] != 0 )
			{
				List *Item1 = Sym.GetItem( CString( Item->Segment ) );
				m_ObjFile->WriteFixupIndex( Item1->Index );
			}
			//Target index
			m_ObjFile->WriteFixupIndex( Index );
			//Fixup end
		}

		if( pmot->Reg_operand_valid )
			MemOp |= pmot->Reg_operand << 3 ;

		MachineCode[0]=pmot->op_code;
		if(pmot->w)
			MachineCode[0]  |= W;
		MachineCode[1] = MemOp;
		if(NOP)
		{
			MachineCode[ret-1]= 0x90;
		}
		MachineCode[ disp + 2 ] = GetLow(ImmOp);
		if(W)
			MachineCode[ disp + 3 ] = GetHigh(ImmOp);
	}
	return ret; //ret;
}

void CAssemble::Error(CString &Operand)
{
	Errors++;
	printf("\nError: (%d) ",LineNo);
	Operand.Print();
	#ifdef WAIT
	getch();
	fflush(stdin);
	#endif
}

void CAssemble::Error(char *Operand)
{
	Errors++;
	printf("\nError: (%d) ",LineNo);
	if( Operand )
		printf("%s",Operand);
	else
		printf("NULL");
	#ifdef WAIT
	getch();
	fflush(stdin);
	#endif
}

void CAssemble::Warning(CString &Operand)
{
	Warnings++;
	printf("\nWarning: (%d) ",LineNo);
	Operand.Print();
}

void CAssemble::Warning(char *Operand)
{
	Warnings++;
	printf("\nWarning: (%d) ",LineNo);
	if( Operand )
		printf("%s",Operand);
	else
		printf("NULL");
}

void CAssemble::WriteSymbolTable(FILE *fp)
{
	List *Item;
	int q;
	//Segments
	fprintf(fp,"\nSymbol Table");
	fprintf(fp,"\n\n-------------------------------------------");
	fprintf(fp,"\nSegments & Groups Alignment  Combine Class");
	fprintf(fp,"\n-------------------------------------------");

	for(q=0;;q++)
	{
		Item=Sym.GetAt(q);

		if(Item==NULL)
			break;
		if( Item->Type != 'S' )
			continue;
		fprintf(fp,"\n%-16s\t",Item->symbol);
		switch( Item->Align )
		{
			case 0:
				fprintf(fp,"Absolute");
				break;
			case 1:
				fprintf(fp,"Byte    ");
				break;
			case 2:
				fprintf(fp,"Word    ");
				break;
			case 3:
				fprintf(fp,"Para    ");
				break;
			case 4:
				fprintf(fp,"Page    ");
				break;
		}
		// 	P-public, S-Stack ' 'Private
		switch( Item->CombineClass )
		{
			case ' ':
				fprintf(fp," Private ");
				break;
			case 'P':
				fprintf(fp," Public  ");
				break;
			case 'S':
				fprintf(fp," Stack   ");
				break;
		}
	}
	fprintf(fp,"\n-------------------------------------------");

	fprintf(fp,"\n\n-----------------------------------------------------------");
	fprintf(fp,"\nSymbol           Type       Value  ");
	fprintf(fp,"\n-----------------------------------------------------------");

	for(q=0;;q++)
	{
		Item=Sym.GetAt(q);

		if(Item==NULL)
			break;
		if( Item->Type == 'S' )
			continue;
		fprintf(fp,"\n%-16s",Item->symbol);
		//Item->Type
		//N-Near, F-Far, B-Byte, W-Word, D-Double A-Absolute
		switch( Item->Type )
		{
			case 'N':
				fprintf(fp," Near     "); break;
			case 'F':
				fprintf(fp," Far      "); break;
			case 'B':
				fprintf(fp," Byte     "); break;
			case 'W':
				fprintf(fp," Word     "); break;
			case 'D':
				fprintf(fp," Double   "); break;
			case 'A':
				fprintf(fp," Absolute "); break;
		}
		fprintf(fp,"\t%s:%0.4X ",Item->Segment,Item->value );
		switch( Item->CombineClass )
		{
			case 'E':
				fprintf(fp," Extern  "); break;
			case 'P':
				fprintf(fp," Public  "); break;
			case ' ':
				fprintf(fp," Private "); break;
		}

	}
	fprintf(fp,"\n-----------------------------------------------------------");

}

BYTE CAssemble::GetHigh(int value)
{
	BYTE temp;
	asm{
			mov ax, value
			mov	temp, ah
		}
	return temp;
}

BYTE CAssemble::GetLow(int value)
{
	BYTE temp;
	asm{
			mov ax, value
			mov	temp, al
		}
	return temp;
}

int CAssemble::GetAddress(CString Operand,int &error)
{
	int disp;
	List * Item;
	error=0;
	if( CanBeLabel(Operand) )
	{
		Item = Sym.GetItem(Operand);
		if(Item == NULL)
		{
			Error("Label undefined");
			Operand.Print("- %s");
			error=1;
			return 0;
		}
		if(Item->Relocation!='R' )
		{
			Error("Invalid Operand");
			error=1;
			return 0;
		}
		disp=Item->value;
	}
	else
	{
		Operand= StripBraces(Operand);
		if( ! IsImmediate(Operand) )
			return 0;
		disp=Operand.ToInteger();
	}
	return disp;
}

BYTE CAssemble::CheckIndices(CString &Part1, CString &Part2, int &disp, int pass, int &error)
{
	int R1, R2, Wide, Seg, R;
	CString Reg1, Reg2;
	BYTE MemOp = 0 ;
	error = 0;

		Reg1 = StripBraces(Part1);
		if( (R1=Tables.IsRegister(Reg1,Wide,Seg))>=0 )
		{
				if( !Wide || Seg )
				{
					Error("Invalid index combination");
					error = 1;
					return 0;
				}
				if( Part2.GetLength() ) 	//two indices
				{
					Reg2 = StripBraces(Part2);
					R2=Tables.IsRegister(Reg2,Wide,Seg);
					if( !Wide || Seg )
					{
						Error("Invalid index combination");
						error = 1;
						return 0;
					}
					if( ( ( R1 == 3 || R1 == 5 ) && ( R2== 6 || R2== 7 ) )
						 || ( ( R2 == 3 || R2 == 5 ) && ( R1== 6 || R1== 7 ) )  )
					{
						if(pass==2)
						{
							if(disp==1)//displacement 1byte
								MemOp=64; //Mod = 2^6
							else if(disp==2)
								MemOp=128; //Mod = 2^7
							else        //No displacement
								MemOp=0;
							//Add R/M bits to MemOp
							R=R1+R2;
							if(R==9);		//BX+SI
							//  MemOp |=0;
							else if (R==10) //BX+DI
								MemOp |=1;
							else if (R==11)	//BP+SI
								MemOp |= 2;
							else			//BD+DI
								MemOp |= 3;
						}
					}
					else
					{
						Error("Invalid index combination");
						error = 1;
						return 0;
					}
				}
				else //only one index
				{

					if( R1==3 || R1==7 || R1==6 || R1==5)
					{
						if(pass==2)
						{
							if(disp==1)//displacement 1byte
								MemOp=64; //Mod = 2^6
							else if(disp==2)
								MemOp=128; //Mod = 2^7
							else        //No displacement
							{
								if(R1==5)
								{
									MemOp = 64;
									disp = 1; //BP instruction displacement is always zero
								}
								else
									MemOp = 0;
							}
							//Add R/M bits to MemOp
							if(R1==3)	 	 //BX
								MemOp |= 7;
							else if (R1== 5) //BP
								MemOp |= 6;
							else if (R1==6)  //SI
								MemOp |= 4;
							else 			 //DI 7
								MemOp |= 5;

						}
					}
					else
					{
						Error("Invalid index combination");
						error = 1;
						return 0;
					}
				}
		}
		return MemOp;
}