#include <stdlib.h>
#include "obj.h"

CObjFile::CObjFile()
{
	Error("\nError:  Class CObjFile defined without parameters");
}

CObjFile::CObjFile(CString &File, CString &Name)
{
	CString Copyright="Copyright (C) 2000 Manoj Paul Joseph";
	BYTE Len = Name.GetLength();
	BYTE CommentClass = 1;
	BYTE Attrib = 0x80;
	Chksum = 0;
	Length = 0;
	HeaderWritten = 0;
	Code = 0;
	Segment = 0;
	Offset = 0;
	RelativeOffset = 0;
	LengthLocation = 0;
	FixStart = NULL;
	FixLast = NULL;
	fp = File.FileOpen("wb");

	//Writing THEADR
	WriteRecHdr(0x80);
	WriteRec( &Len ,1);
	WriteRec(Name.GetPointer(), Name.GetLength() );
	WriteChksum();

	//Writing Comment
	WriteRecHdr(0x88);
	WriteRec( &Attrib ,1);
	WriteRec( &CommentClass ,1);
	WriteRec(Copyright.GetPointer(), Copyright.GetLength() );
	WriteChksum();
}

CObjFile::~CObjFile()
{
	BYTE ModuleType = 0 ; //No start address
	//MODEND Record
	if( Code )
		FlushCode();
	if( FixStart )
		FlushFixup();
	WriteRecHdr(0x8A);
	WriteRec( &ModuleType ,1);
	WriteChksum();

	fclose( fp );
}

BOOL CObjFile::WriteRecHdr(BYTE Header)
{
	int dummy = 0;
	if( Code )
		FlushCode();
	if( FixStart  && Header != 0x9C && Header != 0xA0 && Header != 0xA2 )
		FlushFixup();

	if(HeaderWritten != 0)
	{
		Error("Error writing Record header");
		return 0;
	}
	HeaderWritten = 1;
	Chksum = Header;
	Length = 0;
	fwrite(&Header, 1, 1, fp);
	LengthLocation = ftell(fp);
	fwrite(&dummy, 2, 1, fp);
	return 1;
}

BOOL CObjFile::WriteRec(const BYTE* Start, int length)
{
	if( HeaderWritten == 0 ) //Header not written
	{
		Error("Error writing Record");
		return 0;
	}
	fwrite(Start, 1, length, fp);
	for( int i=0; i< length; i++)
		Chksum += *(Start+i);
	Length += length;
	return 1;
}

BOOL CObjFile::WriteChksum()
{
	if(HeaderWritten == 0)
	{
		Error("Error writing Chksum");
		return 0;
	}

	fseek(fp, LengthLocation, SEEK_SET);
	Length ++;
	fwrite(&Length, 2, 1, fp);
	Chksum = - (Chksum + Length);
	fseek(fp, 0, SEEK_END);

	fwrite(&Chksum, 1, 1, fp);
	Length = 0;
	Chksum = 0;
	HeaderWritten = 0;
	LengthLocation = 0;
	return 1;
}

BOOL CObjFile::WriteLNameHdr()
{
	return WriteRecHdr(0x96);
}

BOOL CObjFile::WriteLName(CString &LName)
{
	BOOL ret;
	BYTE Len = LName.GetLength();
	ret = WriteRec( &Len, 1 );
	ret &= WriteRec( LName.GetPointer(), LName.GetLength() );
	return ret;
}

BOOL CObjFile::WriteLName()
{
	return WriteChksum();

}

BOOL CObjFile::WriteIndex(unsigned int idx)
{
	BYTE Indexb;
	WORD Indexw;
	if(idx <= 0x7F )
	{
		Indexb = idx;
		return WriteRec( &Indexb, 1 );
	}
	else if( idx <= 0x7FFF )
	{
		Indexw = idx;
		Indexw |= 0x8000;
		return WriteRec( (const unsigned char * ) &Indexw, 2 );
	}
	else
		Error("Error writing index");
	return 0;
}

BOOL CObjFile::WriteFixupIndex(unsigned int idx)
{
	BYTE Indexb;
	WORD Indexw;
	if(idx <= 0x7F )
	{
		Indexb = idx;
		return WriteFixupRec( &Indexb, 1 );
	}
	else if( idx <= 0x7FFF )
	{
		Indexw = idx;
		Indexw |= 0x8000;
		return WriteFixupRec( (const unsigned char * ) &Indexw, 2 );
	}
	else
		Error("Error writing index");
	return 0;
}

BOOL CObjFile::AcceptCodeHdr(WORD Seg, WORD Off)
{
	if( Code )
		FlushCode();
	Segment = Seg;
	RelativeOffset = 0;
	Offset = Off;
	return 1;
}

BOOL CObjFile::WriteCode(BYTE const * Start, int length )
{
	if( RelativeOffset + length >= 1024 )
	{
		FlushCode();
	}
	if( Code == 0 )
	{
		//Record header
		WriteRecHdr(0xA0);
		WriteIndex( Segment );
		WriteRec( (const BYTE *) &Offset, 2);
		Code = 1;
	}
	Offset += length;
	RelativeOffset += length;
	return WriteRec( (const BYTE *) Start, length);
}

BOOL CObjFile::WriteCodeData(void * Start, int length )
{
	if( RelativeOffset + length >= 1024 )
	{
		FlushCode();
	}
	if( Code == 0 )
	{
		//Record header
		WriteRecHdr(0xA0);
		WriteIndex( Segment );
		WriteRec( (const BYTE *) &Offset, 2);
		Code = 1;
	}
	Offset += length;
	RelativeOffset += length;
	return WriteRec( (const BYTE *) Start, length);
}


BOOL CObjFile::FlushCode()
{
	int ret = 1;
	if(Code)
	{
		Code=0;
		RelativeOffset = 0;
		ret = WriteChksum();
	}
	FlushFixup();
	return ret;
}

BOOL CObjFile::WriteFixupRec(const BYTE * Start, int length)
{
	for( int i = 0; i < length ; i++ )
		AddFixupRec( Start + i );
	return 1;
}

BOOL CObjFile::AddFixupRec(const BYTE * Data)
{
	struct Record * NewLink;
	NewLink = new (struct Record);
	if (NewLink == NULL )
		return 0;

	NewLink->Record = *Data;
	NewLink->Next = NULL;
	if( FixStart == NULL )
	{
		FixStart = NewLink;
		FixLast  = NewLink;
	}
	else
	{
		FixLast->Next = NewLink;
		FixLast = NewLink;
	}
	return 1;
}

BOOL CObjFile::FlushFixup()
{
	struct Record * Temp, *Prev;
	if ( FixStart == NULL )
		return 0;
	WriteRecHdr(0x9C);
	for(Temp = FixStart; Temp != NULL; )
	{
		WriteRec( & Temp->Record , 1);
		Prev = Temp;
		Temp = Temp->Next;
		delete Prev;
	}
	FixStart = NULL;
	FixLast  = NULL;
	WriteChksum();
	return 1;
}

WORD CObjFile::GetRelativeOffset()
{
	return RelativeOffset ;
}

void CObjFile::Error(char *Message)
{
	printf("%s", Message);
	exit(0);
}
